IMPLEMENTATION MODULE Heaps;
  (* Copyright (C) 1989, Rowley Associates
                         32 Rowley
                         Cam
                         Dursley
                         Glos. GL11 5NT
                         England.

     RcsId: $Id: heaps.mod,v 1.1 93/01/26 22:09:18 paul Exp $

     Notes:  A heap is implemented as a linked list of free blocks.
             These blocks are ordered according to their addresses; free
             blocks with low addresses are placed at the front of the free
             list, and blocks with higher addresses are at the end of the
             free list.

  *)

  FROM SYSTEM IMPORT
    ADDADR, DIFADR, ADDRESS;
  FROM Memory IMPORT
    StaticHeap, GetMemory;

  TYPE
    HeapPtr = POINTER TO HeapRec;
    HeapRec = RECORD
                next: HeapPtr;
                size: LONGCARD
              END;

  CONST
    DefaultHeapSize    = 17*1024;  (* 17K for default heap *)
    AdditionalHeapSize = 32*1024;  (* 32K for each subsequent heap segment *)

  CONST
    NORAM         = 6;
    DOUBLEDISPOSE = 7;

  CONST
    HeapRecSize = SIZE(HeapRec);

  VAR
    More: BOOLEAN;

  PROCEDURE InitHeap(VAR heap: HeapPtr);
    VAR ptr: HeapPtr;
        adr: ADDRESS;
        size: LONGCARD;
  BEGIN
    IF StaticHeap() THEN
      (* Initialise the static heap. *)
      IF More THEN
        GetMemory(adr, size);
        IF (adr = NIL) & (size = 0) THEN
          More := FALSE
        ELSIF size > HeapRecSize*2 THEN
          DEC(size, size MOD HeapRecSize);
          Deallocate(heap, adr, size)
        END
      END
    ELSE
      (* Start off with a default heap. *)
      size := DefaultHeapSize;

      (* round up to a multiple of HeapRecSize. *)
      INC(size, (HeapRecSize - size MOD HeapRecSize) MOD HeapRecSize);

      (* request some memory. *)
      GetMemory(ptr, size);
      IF ptr # NIL THEN

        (* Set up one big free block. *)
        heap := ptr;
        heap^.size := size;
        heap^.next := NIL
      END
    END
  END InitHeap;

  PROCEDURE ExtendHeap(VAR heap: HeapPtr; minSize: LONGCARD; VAR done: BOOLEAN);
    VAR p: HeapPtr;
        adr: ADDRESS;
        size: LONGCARD;
  BEGIN
    IF StaticHeap() THEN
      done := FALSE;
      IF More THEN
        GetMemory(adr, size);
        IF (adr = NIL) & (size = 0) THEN
          More := FALSE;
        ELSIF size > HeapRecSize*2 THEN
          DEC(size, size MOD HeapRecSize);
          Deallocate(heap, adr, size);
          done := TRUE
        END
      END
    ELSE
      (* An additional amount for other requests. *)
      INC(minSize, DefaultHeapSize);

      (* Check for a minimum allocation. *)
      IF minSize < AdditionalHeapSize THEN
        minSize := AdditionalHeapSize
      END;

      (* round up to a multiple of HeapRecSize. *)
      INC(minSize, HeapRecSize - minSize MOD HeapRecSize);

      (* Get some memory from the operating system. *)
      GetMemory(p, minSize);
      done := p # NIL;

      (* Append the new heap segment to the free list, if successful. *)
      IF done THEN
        Deallocate(heap, p, minSize)
      END
    END
  END ExtendHeap;

  PROCEDURE Allocate(VAR heap: HeapPtr; VAR adr: ADDRESS; amount: LONGCARD);
    VAR p, q: HeapPtr;
        done: BOOLEAN;
  BEGIN
    (* If heap wasn't set up by initialisation, try to do it now *)
    IF heap = NIL THEN InitHeap(heap) END;

    (* check for good allocation *)
    IF heap = NIL THEN

      (* still no memory, return NIL *)
      adr := NIL
    ELSE

      (* round up to a multiple of HeapRecSize *)
      IF amount <= HeapRecSize THEN
        amount := HeapRecSize
      ELSE
        INC(amount, (HeapRecSize - amount MOD HeapRecSize) MOD HeapRecSize)
      END;

      (* Find first block that could satisfy request. *)
      p := heap; q := NIL;
      LOOP
        IF p = NIL THEN

          (* Not satisfied, add another heap segment and reallocate. *)
          ExtendHeap(heap, amount, done);
          IF done THEN

            (* More heap was added to satisfy the request, so reallocate. *)
            Allocate(heap, adr, amount)
          ELSE

            (* No more heap space, return NIL. *)
            adr := NIL
          END;

          (* Exit the allocator. *)
          EXIT;

        (* Check whether block is correct size or has to be split. *)
        ELSIF p^.size >= amount THEN
          IF p^.size = amount THEN

            (* Exact fit - remove block from the free list. *)
            adr := p;
            IF q = NIL THEN heap := p^.next
            ELSE q^.next := p^.next
            END
          ELSE

            (* Block must be split; take free memory from top of block. *)
            adr := ADDADR(p, p^.size - amount);
            DEC(p^.size, amount);
          END;

          (* Exit the allocator - block acquired. *)
          EXIT
        ELSE
          q := p; p := p^.next
        END
      END
    END
  END Allocate;

  PROCEDURE Deallocate(VAR heap: HeapPtr; VAR adr: ADDRESS; amount: LONGCARD);
    VAR p, rover, prev: HeapPtr;
        end: ADDRESS;
  BEGIN
    (* Check for dispose of NIL. *)
    IF adr = NIL THEN
      HALT(DOUBLEDISPOSE);
    END;

    (* round up to a multiple of HeapRecSize *)
    IF amount <= HeapRecSize THEN
      amount := HeapRecSize
    ELSE
      INC(amount, (HeapRecSize - amount MOD HeapRecSize) MOD HeapRecSize)
    END;

    (* Deallocated onto non-existent heap? *)
    IF heap = NIL THEN

      (* Yes, add this floating block to the heap. *)
      heap := adr;
      heap^.next := NIL; heap^.size := amount

    ELSE

      (* Search for block adjacent to p, but with lower address. *)
      p := adr; rover := heap; prev := NIL;
      WHILE (rover # NIL) & (DIFADR(rover, adr) < 0) DO
        prev := rover; rover := rover^.next
      END;

      (* Check for double-dispose of a pointer. *)
      (* For efficiency, this could be removed. *)
      IF (rover = adr) OR
         (prev # NIL) & (DIFADR(adr, ADDADR(prev, prev^.size)) < 0) THEN
        HALT(DOUBLEDISPOSE);
      END;

      (* Return block to free list *)
      p^.next := rover; p^.size := amount;
      IF prev = NIL THEN heap := p
      ELSE prev^.next := p
      END;

      (* See if can join with block to right. *)
      IF (rover # NIL) & (ADDADR(p, amount) = rover) THEN
        INC(p^.size, rover^.size);
        p^.next := rover^.next;
      END;

      (* See if can join with block to left. *)
      IF (prev # NIL) & (ADDADR(prev, prev^.size) = p) THEN
        INC(prev^.size, p^.size);
        prev^.next := p^.next;
      END
    END;
    adr := NIL
  END Deallocate;

  PROCEDURE Available(heap: HeapPtr; amount: LONGCARD): BOOLEAN;
    VAR p: HeapPtr;
        done: BOOLEAN;
  BEGIN
    (* find first block that could satisfy request *)
    p := heap;
    WHILE (p # NIL) & (p^.size < amount) DO
      p := p^.next
    END;

    (* return whether request could be satisfied *)
    IF p = NIL THEN

      (* Can't satisfy it from this request.  Try to extend the heap. *)
      ExtendHeap(heap, amount, done);

      (* Return if heap was extended such that request is satisfied. *)
      RETURN done
    ELSE

      (* request could be satisfied from the exiosting heap. *)
      RETURN p^.size >= amount
    END
  END Available;

  PROCEDURE TotalHeapAvail(heap: HeapPtr): LONGCARD;
    VAR total: LONGCARD;
  BEGIN
    total := 0;
    WHILE heap # NIL DO
      INC(total, heap^.size);
      heap := heap^.next;
    END;
    RETURN total
  END TotalHeapAvail;

BEGIN
  More := TRUE
END Heaps.
